<?php

/**
 * @file
 * Contains API documentation and examples for the Paragraphs access module.
 */

use Drupal\paragraphs\ParagraphInterface;
use Drupal\Core\Session\AccountInterface;

/**
 * @addtogroup hooks
 * @{
 */

/**
 * Provide grants for a given user for paragraph access.
 *
 * Can return a list of grants for a given module. Grants are in the format
 * `"realm" => array(...$gids)`. Multiple realms and grants can be supplied by
 * each instance of the hook.
 *
 * @param string $operation
 *   Entity Operation for grants to apply to. one of edit, update, delete.
 * @param \Drupal\Core\Session\AccountInterface $account
 *   Account to build grants for.
 *
 * @return array
 *   mutli value nested array, where keys are realm ids, and values are lists
 *   of grants permitted for a user in that realm.
 */
function hook_paragraphs_access_grants($operation, AccountInterface $account) {
  if ($account === $somecondition) {
    // For accounts that have $somecondition grant 1, 2, 3 in realm_1 and only
    // 1, 2 in realm_2.
    return [
      "realm_1" => [1, 2, 3],
      "realm_2" => [1, 2],
    ];
  }
}

/**
 * Provide grants for a paragraph item.
 *
 * Grants apply to a realm, and give specific access to an item. Grant can
 * provide view, update, or delete access.
 *
 * @param \Drupal\paragraphs\ParagraphInterface $paragraph
 *   A paragraph to build grants for.
 *
 * @return array
 *   Array of grants and the access each provides for the user.
 */
function hook_paragraphs_access_records(ParagraphInterface $paragraph) {
  if ($somecondition) {
    // For paragraphs with a condition, build grants.
    $grants = [];

    // For 2 realms, make 3 grants.
    $realms = ["realm_1", "realm_2"];
    foreach ($realms as $realm) {
      // Grant 1, only provides view.
      $grants[] = [
        'realm' => $realm,
        'gid' => 1,
        'grant_view' => 1,
        'grant_update' => 0,
        'grant_delete' => 0,
        'priority' => 0,
      ];
      // Grant 2, provides view and edit. Not delete.
      $grants[] = [
        'realm' => $realm,
        'gid' => 2,
        'grant_view' => 1,
        'grant_update' => 1,
        'grant_delete' => 0,
        'priority' => 0,
      ];
      // Grant 3, provides view, edit and delete.
      $grants[] = [
        'realm' => $realm,
        'gid' => 3,
        'grant_view' => 1,
        'grant_update' => 1,
        'grant_delete' => 1,
        'priority' => 0,
      ];
    }
    return $grants;
  }
}

/**
 * Alter the message that is displayed for paragraphs the user can't edit.
 *
 * @param \Drupal\paragraphs\ParagraphInterface $paragraph
 *   Item that the users denied access to.
 * @param string $operation
 *   Entity Operation for grants to apply to. one of edit, update, delete.
 * @param array $message
 *   Current render element markup.
 */
function hook_paragraphs_access_restriction_message_alter(ParagraphInterface $paragraph, $operation, array &$message) {
  // Set message on blocked items to something polite.
  $message["#markup"] = "Sorry, You may not " . $operation . " this item.";
}

/**
 * @}
 */
