<?php

namespace Drupal\paragraphs_access;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Query\Condition;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Session\AccountInterface;
use Drupal\paragraphs\ParagraphInterface;
use Drupal\paragraphs\Entity\Paragraph;

/**
 * Defines a storage handler for the paragraphs access grants system.
 *
 * This is used to build paragraphs access queries.
 *
 * @ingroup paragraphs_access
 */
class ParagraphsAccessStorage {

  use StringTranslationTrait;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The language manager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * Constructs a ParagraphsAccessStorage object.
   *
   * @param \Drupal\Core\Database\Connection $database
   *   The database connection.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   * @param \Drupal\Core\Language\LanguageManagerInterface $language_manager
   *   The language manager.
   */
  public function __construct(Connection $database, ModuleHandlerInterface $module_handler, LanguageManagerInterface $language_manager) {
    $this->database = $database;
    $this->moduleHandler = $module_handler;
    $this->languageManager = $language_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function checkGrants(ParagraphInterface $entity, $operation, AccountInterface $account) {

    // Grants only support these operations.
    if (!in_array($operation, ['view', 'update', 'delete'])) {
      return AccessResult::neutral();
    }

    if ($this->hasBypassPermission($account)) {
      return AccessResult::allowed();
    }

    // If no module implements the hook don't query the database for grants.
    if (!$this->moduleHandler->getImplementations('paragraphs_access_grants')) {
      // Return the equivalent of the default grant, defined by
      // self::writeDefault().
      if ($operation === 'view') {
        return AccessResult::allowed()->addCacheableDependency($entity);
      }
      else {
        return AccessResult::neutral();
      }
    }

    if ($operation === "delete") {
      return $this->checkGrants($entity, "update", $account);
    }

    // Check the database for potential access grants.
    $query = $this->database->select('paragraphs_access', 'base');
    $query->addExpression('1');
    // Only interested for granting in the current operation.
    $query->where('grant_' . $operation . '>= 1');
    $grants = $this->getUserGrants($operation, $account);

    // Return rows for the grants the user has.
    $cond = new Condition('OR');
    if (count($grants)) {
      // Check for grants for this node and the correct langcode.
      $id = $query->andConditionGroup()
        ->condition('base.entity_id', $entity->id())
        ->condition('langcode', $entity->language()->getId());
      $id->condition(static::buildAccessCondition($grants));
      $cond->condition($id);
    }
    // If there are no grants, we want to return the default (`all`) row.
    $cond->condition(static::defaultCondition($entity));
    $query->condition($cond);
    $query->range(0, 1);

    if ($query->execute()->fetchField()) {
      $access_result = AccessResult::allowed();
    }
    else {
      $access_result = AccessResult::forbidden();
    }

    // Only the 'view' node grant can currently be cached; the others currently
    // don't have any cacheability metadata. Hopefully, we can add that in the
    // future, which would allow this access check result to be cacheable in all
    // cases. For now, this must remain marked as uncacheable, even when it is
    // theoretically cacheable, because we don't have the necessary metadata to
    // know it for a fact.
    $access_result->addCacheContexts(['user.paragraphs_access:' . $operation]);
    // If ($operation !== 'view') { // disable Caching of result.
    $access_result->setCacheMaxAge(0);
    // }.
    return $access_result;
  }

  /**
   * Check if a given user has the permission to bypass paragraph access checks.
   */
  public function hasBypassPermission(AccountInterface $user) {
    return $user->hasPermission('bypass paragraphs access');
  }

  /**
   * Empty records storage table in database.
   */
  public function deleteRecords() {
    $this->database->truncate('paragraphs_access')->execute();
  }

  /**
   * Generate and insert default grant into database.
   */
  public function saveDefaultGrant() {
    $fields = [
      'entity_id',
      'langcode',
      'fallback',
      'realm',
      'gid',
      'grant_view',
      'grant_update',
      'grant_delete',
    ];
    $query = $this->database->insert('paragraphs_access')->fields($fields);
    $grant = [
      "entity_id" => 0,
      "langcode" => 'en',
      "fallback" => 1,
      "realm" => 'all',
      "gid" => 0,
      "grant_view" => 1,
      "grant_update" => 1,
      "grant_delete" => 1,
    ];
    $query->values($grant);
    $query->execute();
  }

  /**
   * Saved records for a paragraph to the database.
   */
  public function saveRecords(ParagraphInterface $paragraph, array $grants, $delete = TRUE) {
    if ($delete) {
      $query = $this->database->delete('paragraphs_access')->condition('entity_id', $paragraph->id());
      $query->execute();
    }
    // Only perform work when node_access modules are active.
    if (!empty($grants) && count($this->moduleHandler->getImplementations('paragraphs_access_grants'))) {
      $fields = [
        'entity_id',
        'langcode',
        'fallback',
        'realm',
        'gid',
        'grant_view',
        'grant_update',
        'grant_delete',
      ];
      $query = $this->database->insert('paragraphs_access')->fields($fields);
      // If we have defined a granted langcode, use it. But if not, add a grant
      // for every language this paragraph is translated to.
      foreach ($grants as $grant) {
        if (isset($grant['langcode'])) {
          $grant_languages = [$grant['langcode'] => $this->languageManager->getLanguage($grant['langcode'])];
        }
        else {
          $grant_languages = $paragraph->getTranslationLanguages(TRUE);
        }
        foreach ($grant_languages as $grant_langcode => $grant_language) {
          // Only write grants; denies are implicit.
          if ($grant['grant_view'] || $grant['grant_update'] || $grant['grant_delete']) {
            $grant['entity_id'] = $paragraph->id();
            $grant['langcode'] = $grant_langcode;
            // The record with the original langcode is used as the fallback.
            if ($grant['langcode'] == $paragraph->language()->getId()) {
              $grant['fallback'] = 1;
            }
            else {
              $grant['fallback'] = 0;
            }
            $query->values($grant);
          }
        }
      }
      $query->execute();
    }
  }

  /**
   * Reload all access paragraph items.
   *
   * @param bool $batch_mode
   *   If true, create and execute rebuild as a batch job.
   */
  public function reloadRecords($batch_mode = FALSE) {
    if ($batch_mode) {
      $batch = [
        'title' => $this->t('Rebuilding Paragraphs Access Permissions'),
        'operations' => [
          ['_paragraphs_access_rebuild_batch_operation', []],
        ],
        'finished' => '_paragraphs_access_rebuild_batch_finished',
      ];
      batch_set($batch);
    }
    else {
      // Try to allocate enough time to rebuild node grants.
      drupal_set_time_limit(240);

      $paragraph_storage = \Drupal::entityManager()->getStorage('paragraph');
      $this->deleteRecords();
      $this->saveDefaultGrant();

      $entity_query = \Drupal::entityQuery('paragraph');
      $entity_query->sort('entity_id', 'DESC');

      // Disable access checking since all entries must be processed even if the
      // user does not have access.
      $entity_query->accessCheck(FALSE);
      $ids = $entity_query->execute();

      foreach ($ids as $id) {
        $grants = [];
        $paragraph_storage->resetCache([$id]);
        $paragraph = Paragraph::load($id);
        // Only insert and save records for valid nodes.
        if (!empty($paragraph)) {
          $grants = $this->getRecordsFor($paragraph);
          $this->saveRecords($paragraph, $grants);
        }
      }
    }

    if (!isset($batch)) {
      drupal_set_message($this->t('Paragraphs access permissions have been rebuilt.'));
      paragraphs_access_needs_rebuild(FALSE);
    }
  }

  /**
   * Provide hook to generate record data for a given paragraph.
   */
  public function getRecordsFor(ParagraphInterface $paragraph) {
    $records = [];
    if ($this->moduleHandler->getImplementations('paragraphs_access_grants')) {
      if ($this->moduleHandler->getImplementations('paragraphs_access_records')) {
        $records = \Drupal::moduleHandler()->invokeAll('paragraphs_access_records', [$paragraph]);
      }
    }
    return $records;
  }

  /**
   * Get access grants for a user (and op)
   */
  public function getUserGrants($op, AccountInterface $account) {
    $grants = [];
    if ($this->moduleHandler->getImplementations('paragraphs_access_grants')) {
      $grants = \Drupal::moduleHandler()->invokeAll('paragraphs_access_grants', [$op, $account]);
    }
    return $grants;
  }

  /**
   * Build default entry conditional for a paragraph.
   */
  protected function defaultCondition(ParagraphInterface $paragraph) {
    $cond = new Condition("AND");

    $query = $this->database->select('paragraphs_access', 'pa');
    $query->addExpression('1');
    $query->condition('pa.entity_id', $paragraph->id());

    $cond->notExists($query);
    $cond->where('base.entity_id = 0');

    return $cond;
  }

  /**
   * Build access condition for query.
   */
  protected function buildAccessCondition(array $paragraphs_access_grants) {
    $cond = new Condition("AND");

    $query = $this->database->select('paragraphs_access', 'pa');
    $query->addField('pa', 'entity_id');
    $query->where('pa.entity_id = base.entity_id');
    $query->condition(static::buildGrantsCondition($paragraphs_access_grants));

    $cond->exists($query);

    return $cond;
  }

  /**
   * Build grant condition set for grants.
   *
   * @param array $grants
   *   List of grants where 'realm' => array(...$gids)
   */
  protected function buildGrantsCondition(array $grants) {
    $cond = new Condition("OR");
    foreach ($grants as $realm => $gids) {
      if (!empty($gids)) {
        $and = new Condition('AND');
        $cond->condition($and
        // ->condition('gid', $gids, 'IN')
          ->condition('realm', $realm)
        );
      }
    }
    return $cond;
  }

  /**
   * Count the number of rows in the table.
   */
  public function count() {
    return $this->database->query('SELECT COUNT(*) FROM {paragraphs_access}')->fetchField();
  }

}
